from .code import CodeSearchResult
from .issue import IssueSearchResult
from .repository import RepositorySearchResult
from .user import UserSearchResult


__all__ = [CodeSearchResult, IssueSearchResult, RepositorySearchResult,
           UserSearchResult]
